﻿// -*- C++ -*-
/*!
 * @file  SimDataPublish.cpp
 * @brief ModuleDescription
 * @date $Date$
 *
 * $Id$
 */

#include "SimDataPublish.h"
#include <chrono>

const double defaultDouble = 1.79769e+308;
const double defaultLong = 2147483647;

// Module specification
// <rtc-template block="module_spec">
static const char* simdataPublish_spec[] =
  {
    "implementation_id", "SimDataPublish",
    "type_name",         "SimDataPublish",
    "description",       "ModuleDescription",
    "version",           "1.0.0",
    "vendor",            "Fukushima Computer System",
    "category",          "Simulator",
//    "activity_type",     "EVENTDRIVEN",
    "activity_type",     "PERIODIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    ""
  };
// </rtc-template>

/*!
 * @brief constructor
 * @param manager Maneger Object
 */
SimDataPublish::SimDataPublish(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : BodyIoRTC(manager),
    m_inAllDataIn("inAllData", m_inAllData),
    m_outAllDataOut("outAllData", m_outAllData)

    // </rtc-template>
{
}

/*!
 * @brief destructor
 */
SimDataPublish::~SimDataPublish()
{
}

void SimDataPublish::GetTime(RTC::Time &tm)
{
    std::chrono::system_clock::time_point now = std::chrono::system_clock::now();
    tm.sec = std::chrono::duration_cast<std::chrono::seconds>(now.time_since_epoch()).count();
    tm.nsec= std::chrono::duration_cast<std::chrono::nanoseconds>(now.time_since_epoch()).count()%1000000000;
//      timespec tspec;
//    timespec_get(&tspec, TIME_UTC);
//      clock_gettime(CLOCK_REALTIME, &tspec);
//      tm.sec = tspec.tv_sec;
//      tm.nsec = tspec.tv_nsec;

//    std::cout << "Unix時間（秒）: " << tm.sec << std::endl;
//    std::cout << "Unix時間（ナノ秒）: " << tm.nsec << std::endl;

}

RTC::ReturnCode_t SimDataPublish::onInitialize(Body* body)
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("inAllData", m_inAllDataIn);

  // Set OutPort buffer
  addOutPort("outAllData", m_outAllDataOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

  // <rtc-template block="bind_config">
  // </rtc-template>
  
  RTC::ExecutionContext_var ec = get_context(0);
  
  // (5Hz)
  ec->set_rate(5);
  

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t SimDataPublish::onFinalize()
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onActivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

void SimDataPublish::dataInitialize(void)
{
    GetTime(m_outAllData.tm);

    GetTime(m_outAllData.robotid.tm);
    m_outAllData.robotid.data = "spider2020_simu_01";

    GetTime(m_outAllData.pose3d.tm);
	m_outAllData.pose3d.data.position.x = defaultDouble;
	m_outAllData.pose3d.data.position.y = defaultDouble;
	m_outAllData.pose3d.data.position.z = defaultDouble;
	m_outAllData.pose3d.data.orientation.r = defaultDouble;
	m_outAllData.pose3d.data.orientation.p = defaultDouble;
	m_outAllData.pose3d.data.orientation.y = defaultDouble;

    GetTime(m_outAllData.motorLoad.tm);
    m_outAllData.motorLoad.data.length(6);
    m_outAllData.motorLoad.data[0] = defaultDouble;
    m_outAllData.motorLoad.data[1] = defaultDouble;
    m_outAllData.motorLoad.data[2] = defaultDouble;
    m_outAllData.motorLoad.data[3] = defaultDouble;
    m_outAllData.motorLoad.data[4] = defaultDouble;
    m_outAllData.motorLoad.data[5] = defaultDouble;

    GetTime(m_outAllData.flipperAngle.tm);
    m_outAllData.flipperAngle.data.length(4);
    m_outAllData.flipperAngle.data[0] = defaultDouble;
    m_outAllData.flipperAngle.data[1] = defaultDouble;
    m_outAllData.flipperAngle.data[2] = defaultDouble;
    m_outAllData.flipperAngle.data[3] = defaultDouble;

    GetTime(m_outAllData.imuOri3d.tm);
    m_outAllData.imuOri3d.data.r = defaultDouble;
    m_outAllData.imuOri3d.data.p = defaultDouble;
    m_outAllData.imuOri3d.data.y = defaultDouble;

    GetTime(m_outAllData.velo2d.tm);
    m_outAllData.velo2d.data.vx = defaultDouble;
    m_outAllData.velo2d.data.vy = defaultDouble;
    m_outAllData.velo2d.data.va = defaultDouble;

    GetTime(m_outAllData.gcCapaRatio.tm);
    m_outAllData.gcCapaRatio.data = defaultDouble;

    GetTime(m_outAllData.gcAxis.tm);
    m_outAllData.gcAxis.data.length(8);
    m_outAllData.gcAxis.data[0] = defaultDouble;
    m_outAllData.gcAxis.data[1] = defaultDouble;
    m_outAllData.gcAxis.data[2] = defaultDouble;
    m_outAllData.gcAxis.data[3] = defaultDouble;
    m_outAllData.gcAxis.data[4] = defaultDouble;
    m_outAllData.gcAxis.data[5] = defaultDouble;
    m_outAllData.gcAxis.data[6] = defaultDouble;
    m_outAllData.gcAxis.data[7] = defaultDouble;

    GetTime(m_outAllData.gcButton.tm);
    m_outAllData.gcButton.data = defaultLong;

    GetTime(m_outAllData.mainBattCapaRatio.tm);
    m_outAllData.mainBattCapaRatio.data = defaultDouble;

    GetTime(m_outAllData.mainSupplyVolt.tm);
    m_outAllData.mainSupplyVolt.data = defaultDouble;

    GetTime(m_outAllData.mainRawSupplyVolt.tm);
    m_outAllData.mainRawSupplyVolt.data = defaultDouble;

    GetTime(m_outAllData.mode.tm);
    m_outAllData.mode.data.length(9);
    m_outAllData.mode.data[0] = defaultLong;
    m_outAllData.mode.data[1] = defaultLong;
    m_outAllData.mode.data[2] = defaultLong;
    m_outAllData.mode.data[3] = defaultLong;
    m_outAllData.mode.data[4] = defaultLong;
    m_outAllData.mode.data[5] = defaultLong;
    m_outAllData.mode.data[6] = defaultLong;
    m_outAllData.mode.data[7] = defaultLong;
    m_outAllData.mode.data[8] = defaultLong;

    GetTime(m_outAllData.arm1Pose3d.tm);
	m_outAllData.arm1Pose3d.data.position.x = defaultDouble;
	m_outAllData.arm1Pose3d.data.position.y = defaultDouble;
	m_outAllData.arm1Pose3d.data.position.z = defaultDouble;
	m_outAllData.arm1Pose3d.data.orientation.r = defaultDouble;
	m_outAllData.arm1Pose3d.data.orientation.p = defaultDouble;
	m_outAllData.arm1Pose3d.data.orientation.y = defaultDouble;

    GetTime(m_outAllData.arm1Angular.tm);
    m_outAllData.arm1Angular.data.length(6);
    m_outAllData.arm1Angular.data[0] = defaultDouble;
    m_outAllData.arm1Angular.data[1] = defaultDouble;
    m_outAllData.arm1Angular.data[2] = defaultDouble;
    m_outAllData.arm1Angular.data[3] = defaultDouble;
    m_outAllData.arm1Angular.data[4] = defaultDouble;
    m_outAllData.arm1Angular.data[5] = defaultDouble;

    GetTime(m_outAllData.arm1Finger.tm);
    m_outAllData.arm1Finger.data.length(3);
    m_outAllData.arm1Finger.data[0] = defaultLong;
    m_outAllData.arm1Finger.data[1] = defaultLong;
    m_outAllData.arm1Finger.data[2] = defaultLong;

    GetTime(m_outAllData.arm1Current.tm);
    m_outAllData.arm1Current.data.length(9);
    m_outAllData.arm1Current.data[0] = defaultDouble;
    m_outAllData.arm1Current.data[1] = defaultDouble;
    m_outAllData.arm1Current.data[2] = defaultDouble;
    m_outAllData.arm1Current.data[3] = defaultDouble;
    m_outAllData.arm1Current.data[4] = defaultDouble;
    m_outAllData.arm1Current.data[5] = defaultDouble;
    m_outAllData.arm1Current.data[6] = defaultDouble;
    m_outAllData.arm1Current.data[7] = defaultDouble;
    m_outAllData.arm1Current.data[8] = defaultDouble;
}


RTC::ReturnCode_t SimDataPublish::onExecute(RTC::UniqueId ec_id)
{
    GetTime(m_outAllData.tm);
//    std::cout << "Unix時間: " << m_outAllData.tm.sec << "." << m_outAllData.tm.nsec << std::endl;

    GetTime(m_outAllData.robotid.tm);

    cnoid::Vector3 eular;
#if 0
//for debug
    //std::cout << "time:" << m_outAllData.tm.sec << ", " << m_outAllData.tm.nsec << std::endl;
    double    orir = 0.24;
    double    orip = 0.4;
    double    oriy = 0.0;
    std::cout << orir << ","<< orip << ","<< oriy << " =>> ";
        Matrix3 R;
        SE3 position;
        eular = cnoid::Vector3(orir, orip, oriy);
        Matrix3 rotFrom = cnoid::rotFromRpy(eular);

        Quaternion& q = position.rotation();
  
        double q0, q1, q2, q3;
        EulerAnglesToQuaternion(orir, orip, oriy, q0, q1, q2, q3);
        q.w() = q0;
        q.x() = q1;
        q.y() = q2;
        q.z() = q3;
       R = position.rotation().toRotationMatrix();

//    std::cout << q0 << "," << q1 << "," << q1 << "," << q1 << "," << "("<< rotFrom <<") =>> ";
    std::cout << R << " --- " << rotFrom;

//        toQuaternionEulaerAngles(R, orir, orip, oriy);
        eular = cnoid::rpyFromRot(rotFrom);

    std::cout << eular[0]<< ","<< eular[1] << ","<< eular[2] << "<<>>";
    std::cout << orir << ","<< orip << ","<< oriy << std::endl;
#endif

    if(m_cnoid_Link[LINK_CHASSIS] != nullptr){
        GetTime(m_outAllData.pose3d.tm);
        double orir, orip, oriy;
        SE3 position;
        Vector3 p = m_cnoid_Link[LINK_CHASSIS]->p();
        m_outAllData.pose3d.data.position.x = p.x();
        m_outAllData.pose3d.data.position.y = p.y();
        m_outAllData.pose3d.data.position.z = p.z();

        eular = cnoid::rpyFromRot(m_cnoid_Link[LINK_CHASSIS]->R());
        m_outAllData.pose3d.data.orientation.r = eular[0];
        m_outAllData.pose3d.data.orientation.p = eular[1];
        m_outAllData.pose3d.data.orientation.y = eular[2];
    }

    GetTime(m_outAllData.motorLoad.tm);

    GetTime(m_outAllData.flipperAngle.tm);
    // flipperAngle
    if(m_cnoid_Link[LINK_FR_FLIPPER] != nullptr){
        m_outAllData.flipperAngle.data[0] = m_cnoid_Link[LINK_FR_FLIPPER]->q();
    }
    if(m_cnoid_Link[LINK_FL_FLIPPER] != nullptr){
        m_outAllData.flipperAngle.data[1] = m_cnoid_Link[LINK_FL_FLIPPER]->q();
    }
    if(m_cnoid_Link[LINK_BR_FLIPPER] != nullptr){
        m_outAllData.flipperAngle.data[2] = m_cnoid_Link[LINK_BR_FLIPPER]->q();
    }
    if(m_cnoid_Link[LINK_BL_FLIPPER] != nullptr){
        m_outAllData.flipperAngle.data[3] = m_cnoid_Link[LINK_BL_FLIPPER]->q();
    }

    GetTime(m_outAllData.imuOri3d.tm);
    GetTime(m_outAllData.velo2d.tm);
    GetTime(m_outAllData.gcCapaRatio.tm);
    GetTime(m_outAllData.gcAxis.tm);
    GetTime(m_outAllData.mainBattCapaRatio.tm);
    GetTime(m_outAllData.mainSupplyVolt.tm);
    GetTime(m_outAllData.mainRawSupplyVolt.tm);
    GetTime(m_outAllData.mode.tm);
    GetTime(m_outAllData.arm1Pose3d.tm);

    // arm1Angular;
    GetTime(m_outAllData.arm1Angular.tm);
    if(m_cnoid_Link[LINK_SHOULDER] != nullptr){
        m_outAllData.arm1Angular.data[0] = m_cnoid_Link[LINK_SHOULDER]->q();
    }
    if(m_cnoid_Link[LINK_ARM] != nullptr){
        m_outAllData.arm1Angular.data[1] = m_cnoid_Link[LINK_ARM]->q();
    }
    if(m_cnoid_Link[LINK_FOREARM] != nullptr){
        m_outAllData.arm1Angular.data[2] = m_cnoid_Link[LINK_FOREARM]->q() ;
    }
    if(m_cnoid_Link[LINK_WRIST1] != nullptr){
        m_outAllData.arm1Angular.data[3] = m_cnoid_Link[LINK_WRIST1]->q();
    }
    if(m_cnoid_Link[LINK_WRIST2] != nullptr){
        m_outAllData.arm1Angular.data[4] = m_cnoid_Link[LINK_WRIST2]->q();
    }
    if(m_cnoid_Link[LINK_HAND] != nullptr){
        m_outAllData.arm1Angular.data[5] = m_cnoid_Link[LINK_HAND]->q();
    }

    // arm1Finger
    GetTime(m_outAllData.arm1Finger.tm);
    double data;
    if(m_cnoid_Link[LINK_FINGER1] != nullptr){
        data =  m_cnoid_Link[LINK_FINGER1]->q() ;
        data = data * 6800 / 115 / 3.14 * 180;
        m_outAllData.arm1Finger.data[0] = data;
    }
    if(m_cnoid_Link[LINK_FINGER2] != nullptr){
        data =  m_cnoid_Link[LINK_FINGER2]->q() ;
        data = data * 6800 / 115 / 3.14 * 180;
        m_outAllData.arm1Finger.data[1] = data;
    }
    if(m_cnoid_Link[LINK_FINGER3] != nullptr){
        data =  m_cnoid_Link[LINK_FINGER3]->q() ;
        data = data * 6800 / 115 / 3.14 * 180;
        m_outAllData.arm1Finger.data[2] = data;
    }

    GetTime(m_outAllData.arm1Current.tm);

    m_outAllDataOut.write();

    return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t SimDataPublish::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

/*
RTC::ReturnCode_t SimDataPublish::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
bool SimDataPublish::initializeSimulation(ControllerIO* io)
{
    int i = 0;
    
    if(io != nullptr){
       Body* body = io->body();
       if(body != nullptr){
           m_cnoid_Link[LINK_CHASSIS]    = body->link("CHASSIS");
           m_cnoid_Link[LINK_FR_FLIPPER] = body->link("FR_FLIPPER");
           m_cnoid_Link[LINK_FL_FLIPPER] = body->link("FL_FLIPPER");
           m_cnoid_Link[LINK_BR_FLIPPER] = body->link("BR_FLIPPER");
           m_cnoid_Link[LINK_BL_FLIPPER] = body->link("BL_FLIPPER");
           m_cnoid_Link[LINK_SHOULDER]   = body->link("SHOULDER");
           m_cnoid_Link[LINK_ARM]        = body->link("ARM");
           m_cnoid_Link[LINK_FOREARM]    = body->link("FOREARM");
           m_cnoid_Link[LINK_WRIST1]     = body->link("WRIST1");
           m_cnoid_Link[LINK_WRIST2]     = body->link("WRIST2");
           m_cnoid_Link[LINK_HAND]       = body->link("HAND");
           m_cnoid_Link[LINK_FINGER1]    = body->link("FINGER1");           
           m_cnoid_Link[LINK_FINGER2]    = body->link("FINGER2");
           m_cnoid_Link[LINK_FINGER3]    = body->link("FINGER3");
           for(i = 0; i < LINK_NUM; i++){
               if(m_cnoid_Link[i] == NULL){
                   return false;
               }
           }
           
 /*          m_cnoid_Link[LINK_SHOULDER]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_ARM]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_FOREARM]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_WRIST1]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_WRIST2]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_HAND]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_FINGER1]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_FINGER2]->setActuationMode(Link::JOINT_ANGLE);
           m_cnoid_Link[LINK_FINGER3]->setActuationMode(Link::JOINT_ANGLE);*/
       }
    }
    dataInitialize();

    return true;
}

void SimDataPublish::inputFromSimulator() {
//	printf("SHOULDER:%f\n", m_cnoid_Link[LINK_SHOULDER]->q());
//	printf("ARM     :%f\n", m_cnoid_Link[LINK_ARM]->q());
//	printf("FOREARM :%f\n", m_cnoid_Link[LINK_FOREARM]->q());
//	printf("WRIST1  :%f\n", m_cnoid_Link[LINK_WRIST1]->q());
//	printf("WRIST2  :%f\n", m_cnoid_Link[LINK_WRIST2]->q());
//	printf("HAND    :%f\n", m_cnoid_Link[LINK_HAND]->q());
 }

void SimDataPublish::outputToSimulator() {
#if 1
    if(m_inAllDataIn.isNew()){
        m_inAllDataIn.read();
//        std::cout << "output" << std::endl;
    if(m_cnoid_Link[LINK_CHASSIS] != nullptr){
        double orir, orip, oriy;
        SE3 position;
        Vector3& p = position.translation();
        p.x() = m_inAllData.pose3d.data.position.x;
        p.y() = m_inAllData.pose3d.data.position.y;
        p.z() = m_inAllData.pose3d.data.position.z;
        orir = m_inAllData.pose3d.data.orientation.r;
        orip = m_inAllData.pose3d.data.orientation.p;
        oriy = m_inAllData.pose3d.data.orientation.y;
//        Quaternion& q = position.rotation();
//        double q0, q1, q2, q3;
//        EulerAnglesToQuaternion(orir, orip, oriy, q0, q1, q2, q3);
//        q.w() = q0;
//        q.x() = q1;
//        q.y() = q2;
//        q.z() = q3;
        cnoid::Vector3 eular = cnoid::Vector3(orir, orip, oriy);
        Matrix3 rotFrom = cnoid::rotFromRpy(eular);
    
        m_cnoid_Link[LINK_CHASSIS]->p() = position.translation();
//        m_cnoid_Link[LINK_CHASSIS]->R() = position.rotation().toRotationMatrix();
        m_cnoid_Link[LINK_CHASSIS]->R() = rotFrom;
    }
    

        // flipperAngle
        if(m_inAllData.flipperAngle.data.length() >= 4) {
            if(m_cnoid_Link[LINK_FR_FLIPPER] != nullptr){
                m_inAllData.flipperAngle.data[0] = m_cnoid_Link[LINK_FR_FLIPPER]->q();
            }
            if(m_cnoid_Link[LINK_FL_FLIPPER] != nullptr){
                m_inAllData.flipperAngle.data[1] = m_cnoid_Link[LINK_FL_FLIPPER]->q();
            }
            if(m_cnoid_Link[LINK_BR_FLIPPER] != nullptr){
                m_inAllData.flipperAngle.data[2] = m_cnoid_Link[LINK_BR_FLIPPER]->q();
            }
            if(m_cnoid_Link[LINK_BL_FLIPPER] != nullptr){
                m_inAllData.flipperAngle.data[3] = m_cnoid_Link[LINK_BL_FLIPPER]->q();
            }
        }
        
        // arm1Angular;
        if(m_inAllData.arm1Angular.data.length() >= 6) {
            if(m_cnoid_Link[LINK_SHOULDER] != nullptr){
                m_cnoid_Link[LINK_SHOULDER]->q() = m_inAllData.arm1Angular.data[0];
            }
            if(m_cnoid_Link[LINK_ARM] != nullptr){
                m_cnoid_Link[LINK_ARM]->q() = m_inAllData.arm1Angular.data[1];
            }
            if(m_cnoid_Link[LINK_FOREARM] != nullptr){
                m_cnoid_Link[LINK_FOREARM]->q() = m_inAllData.arm1Angular.data[2];
            }
            if(m_cnoid_Link[LINK_WRIST1] != nullptr){
                m_cnoid_Link[LINK_WRIST1]->q() = m_inAllData.arm1Angular.data[3];
            }
            if(m_cnoid_Link[LINK_WRIST2] != nullptr){
                m_cnoid_Link[LINK_WRIST2]->q() = m_inAllData.arm1Angular.data[4];
            }
            if(m_cnoid_Link[LINK_HAND] != nullptr){
                m_cnoid_Link[LINK_HAND]->q() = m_inAllData.arm1Angular.data[5];
            }
        }
    
        // arm1Finger
        if(m_inAllData.arm1Finger.data.length() >= 3) {
            double data;
            if(m_cnoid_Link[LINK_FINGER1] != nullptr){
                data = m_inAllData.arm1Finger.data[0];
                data = data / 6800 * 115 * 3.14 / 180;
                m_cnoid_Link[LINK_FINGER1]->q() = data;
            }
            if(m_cnoid_Link[LINK_FINGER2] != nullptr){
                data = m_inAllData.arm1Finger.data[1];
                data = data / 6800 * 115 * 3.14 / 180;
                m_cnoid_Link[LINK_FINGER2]->q() = data;
            }
            if(m_cnoid_Link[LINK_FINGER3] != nullptr){
                data = m_inAllData.arm1Finger.data[2];
                data = data / 6800 * 115 * 3.14 / 180;
                m_cnoid_Link[LINK_FINGER3]->q() = data;
            }
        }

    }
#endif
}

void SimDataPublish::stopSimulation() {
}

void SimDataPublish::toQuaternionEulaerAngles(Matrix3 m, double& roll, double& pitch, double& yaw) {
    // to Quaternion
    double w, x, y, z;
    double trace = m(0,0) + m(1,1) + m(2,2);
    if (trace > 0) {
        double s = 0.5 / sqrt(trace + 1.0);
        w = 0.25 / s;
        x = (m(2,1) - m(1,2)) * s;
        y = (m(0,2) - m(2,0)) * s;
        z = (m(1,0) - m(0,1)) * s;
    } else {
        if (m(0,0) > m(1,1) && m(0,0) > m(2,2)) {
            double s = 2.0 * sqrt(1.0 + m(0,0) - m(1,1) - m(2,2));
            w = (m(2,1) - m(1,2)) / s;
            x = 0.25 * s;
            y = (m(0,1) + m(1,0)) / s;
            z = (m(0,2) + m(2,0)) / s;
        } else if (m(1,1) > m(2,2)) {
            double s = 2.0 * sqrt(1.0 + m(1,1) - m(0,0) - m(2,2));
            w = (m(0,2) - m(2,0)) / s;
            x = (m(0,1) + m(1,0)) / s;
            y = 0.25 * s;
            z = (m(1,2) + m(2,1)) / s;
        } else {
            double s = 2.0 * sqrt(1.0 + m(2,2) - m(0,0) - m(1,1));
            w = (m(1,0) - m(0,1)) / s;
            x = (m(0,2) + m(2,0)) / s;
            y = (m(1,2) + m(2,1)) / s;
            z = 0.25 * s;
        }
    }
    // Quaternion to QularAngles
    QuaternionToEulerAngles(x, y, z, w, roll, pitch, yaw);

    return;
}

void SimDataPublish::EulerAnglesToQuaternion(double roll, double pitch, double yaw, double& q0, double& q1, double& q2, double& q3)
{
    double cosRoll = cos(roll / 2.0);
    double sinRoll = sin(roll / 2.0);
    double cosPitch = cos(pitch / 2.0);
    double sinPitch = sin(pitch / 2.0);
    double cosYaw = cos(yaw / 2.0);
    double sinYaw = sin(yaw / 2.0);
    
    q0 = cosRoll * cosPitch * cosYaw + sinRoll * sinPitch * sinYaw;
    q1 = sinRoll * cosPitch * cosYaw + cosRoll * sinPitch * sinYaw;
    q2 = cosRoll * sinPitch * cosYaw + sinRoll * cosPitch * sinYaw;
    q3 = cosRoll * cosPitch * sinYaw - sinRoll * sinPitch * cosYaw;
}

void SimDataPublish::QuaternionToEulerAngles(double x, double y, double z, double w, double& roll, double& pitch, double& yaw )
{
    double sinr_cosp = 2 * (w*x+y*z);
    double cosr_cosp = 1 - 2 * (x*x+y*y);
    pitch = std::atan2(sinr_cosp, cosr_cosp);
    
    double sinp = 2*(w*y - z*x);
    if ( std::abs(sinp) >= 1)
        yaw = std::copysign(M_PI / 2, sinp); // use 90 degrees if out of range
    else
        yaw = std::asin(sinp);
    
    double siny_cosp = 2 * (w*z+x*y);
    double cosy_cosp = 1 - 2 * (y*y+z*z);
    roll = std::atan2(siny_cosp, cosy_cosp);
}

bool SimDataPublish::initializeIO(ControllerIO* io) 
{
    return true;
}




extern "C"
{

  void SimDataPublishInit(RTC::Manager* manager)
  {
    coil::Properties profile(simdataPublish_spec);
    manager->registerFactory(profile,
                             RTC::Create<SimDataPublish>,
                             RTC::Delete<SimDataPublish>);
  }

};


